/*
    This is a romfs file system driver for Windows NT and Windows 2000.
    Copyright (C) 1999, 2000, 2001 Bo Brantn.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

NTSTATUS
FsdLookupFileName (
    IN PFSD_VCB         Vcb,
    IN PUNICODE_STRING  FullFileName,
    IN OUT PULONG       Index,
    IN OUT struct romfs_inode* romfs_inode
    );

NTSTATUS
FsdLookupFileNameInDir (
    IN PFSD_VCB         Vcb,
    IN ULONG            DirIndex,
    IN PUNICODE_STRING  FileName,
    IN OUT PULONG       Index,
    IN OUT struct romfs_inode* romfs_inode
    );

NTSTATUS
FsdCreate (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    PDEVICE_OBJECT      DeviceObject;
    PIRP                Irp;
	PIO_STACK_LOCATION	io_stack;
	NTSTATUS			status;
    PFSD_VCB            Vcb;
    PFSD_FCB            Fcb = 0;
    PFSD_CCB            Ccb;
    ULONG               found_index = 0;
    struct romfs_inode* romfs_inode;
    PLIST_ENTRY         Link;
    PFSD_FCB            TmpFcb;

    DeviceObject = IrpContext->DeviceObject;

    Irp = IrpContext->Irp;

	io_stack = IoGetCurrentIrpStackLocation(Irp);

    if (DeviceObject == FsdGlobalData.DeviceObject)
    {
        KdPrint((DRIVER_NAME ": Create: Create on main device object\n"));

        Irp->IoStatus.Information = FILE_OPENED;
        status = STATUS_SUCCESS;
        IrpContext->Irp->IoStatus.Status = status;
    
        FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
    
        FsdFreeIrpContext(IrpContext);
    
        return status;

    }
    else if (io_stack->FileObject->FileName.Length == 0)
    {
        io_stack->FileObject->FsContext = DeviceObject->DeviceExtension;
                
        Irp->IoStatus.Information = FILE_OPENED;
        status = STATUS_SUCCESS;
        KdPrint((DRIVER_NAME ": Create on volume succeded\n"));
        IrpContext->Irp->IoStatus.Status = status;
    
        FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
    
        FsdFreeIrpContext(IrpContext);
    
        return status;
    }

    Vcb = (PFSD_VCB) DeviceObject->DeviceExtension;

    KdPrint((DRIVER_NAME ": Create: FileName: (%S)\n", io_stack->FileObject->FileName.Buffer));

    Link = Vcb->FcbList.Flink;

    while (Link != &Vcb->FcbList)
    {
        TmpFcb = CONTAINING_RECORD(Link, FSD_FCB, Next);

        if (TmpFcb && TmpFcb->Identifier.Type == FCB)
        {
            KdPrint(("[%s,%#x]\n", TmpFcb->DbgFileName, (ULONG)TmpFcb->IndexNumber.QuadPart));

            if (!RtlCompareUnicodeString(
                &TmpFcb->FileName,
                &io_stack->FileObject->FileName,
                FALSE
                ))
            {
                KdPrint(("Found an old FCB!\n"));
                Fcb = TmpFcb;
                break;
            }
        }
        Link = Link->Flink;
    }

    if (!Fcb)
    {
        romfs_inode = FsdAllocatePool(
            NonPagedPool,
            sizeof(struct romfs_inode) + ROMFS_MAXFN
            );

        status = FsdLookupFileName(
            Vcb,
            &io_stack->FileObject->FileName,
            &found_index,
            romfs_inode
            );
        
        KdPrint((DRIVER_NAME ": Create: FileIndex: (%#x)\n", found_index));
        KdPrint((DRIVER_NAME ": Create: FileIndex: (%#x)\n", be32_to_cpu(found_index) & ROMFH_MASK));
        
        if (NT_SUCCESS(status))
        {
            Fcb = FsdAllocateFcb (
                ((PFSD_VCB)DeviceObject->DeviceExtension),
                &io_stack->FileObject->FileName,
                found_index,
                romfs_inode
                );
        }
        else
        {
            IrpContext->Irp->IoStatus.Status = status;
    
            FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
    
            FsdFreeIrpContext(IrpContext);
    
            return status;
        }
    }

    if (Fcb)
    {
        Ccb = FsdAllocateCcb();

        Fcb->OpenHandleCount++;
        Fcb->ReferenceCount++;

        if (!(be32_to_cpu(Fcb->romfs_inode->next) & ROMFH_DIR))
        {
            Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsPossible;
        }

        io_stack->FileObject->FsContext = (void*)Fcb;
        io_stack->FileObject->FsContext2 = (void*) Ccb;
        io_stack->FileObject->PrivateCacheMap = NULL;
        io_stack->FileObject->SectionObjectPointer = &(Fcb->SectionObject);

        Irp->IoStatus.Information = FILE_OPENED;
        status = STATUS_SUCCESS;
        
        KdPrint((
            DRIVER_NAME ": %s OpenHandleCount: %u ReferenceCount: %u\n",
            Fcb->DbgFileName,
            Fcb->OpenHandleCount,
            Fcb->ReferenceCount
            ));
    }
    else
    {
        Irp->IoStatus.Information = 0;
        status = STATUS_OBJECT_NAME_NOT_FOUND;
        KdPrint((DRIVER_NAME ": Create denied\n"));
    }
    
    IrpContext->Irp->IoStatus.Status = status;
    
    FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
    
    FsdFreeIrpContext(IrpContext);
    
    return status;
}

NTSTATUS
FsdLookupFileName (
    IN PFSD_VCB         Vcb,
    IN PUNICODE_STRING  FullFileName,
    IN OUT PULONG       Index,
    IN OUT struct romfs_inode* romfs_inode
    )
{
    ULONG           RootFileIndex;
    UNICODE_STRING  FileName;
    NTSTATUS        Status;

    RootFileIndex = (ROMFH_SIZE +
            strnlen(Vcb->romfs_super_block->name, ROMFS_MAXFN)
            + 1 + ROMFH_PAD) & ROMFH_MASK;

    RootFileIndex = cpu_to_be32(RootFileIndex);

    if (FullFileName->Length == 0)
    {
        *Index = 0;
        return STATUS_OBJECT_NAME_NOT_FOUND;
    }

    if (FullFileName->Length == 2 && FullFileName->Buffer[0] == L'\\')
    {
        Status = FsdReadInodeByIndex(
            Vcb->TargetDeviceObject,
            RootFileIndex,
            romfs_inode
            );

        if (!NT_SUCCESS(Status))
        {
            *Index = 0;
            return Status;
        }

        *Index = RootFileIndex;

        return STATUS_SUCCESS;
    }

    FileName = *FullFileName;

    FileName.Buffer++;
    FileName.Length -= 2;

    return FsdLookupFileNameInDir (
        Vcb,
        RootFileIndex,
        &FileName,
        Index,
        romfs_inode
    );
}

NTSTATUS
FsdLookupFileNameInDir (
    IN PFSD_VCB         Vcb,
    IN ULONG            DirIndex,
    IN PUNICODE_STRING  FileName,
    IN OUT PULONG       Index,
    IN OUT struct romfs_inode* romfs_inode
    )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    ULONG                   FileIndex;
    USHORT                  InodeFileNameLength;
    UNICODE_STRING          InodeFileName;
    UNICODE_STRING  slash_string;

    InodeFileName.Buffer = NULL;

    __try
    {
        FileIndex = DirIndex;

        if (!(be32_to_cpu(FileIndex) & ROMFH_MASK))
        {
            Status = STATUS_NO_SUCH_FILE;
            __leave;
        }

        while ((be32_to_cpu(FileIndex) & ROMFH_MASK))
        {
            Status = FsdReadInodeByIndex(
                Vcb->TargetDeviceObject,
                FileIndex,
                romfs_inode
                );

            if (!NT_SUCCESS(Status))
            {
                Status = STATUS_NO_SUCH_FILE;
                __leave;
            }

            InodeFileNameLength = strnlen(romfs_inode->name, ROMFS_MAXFN);

            InodeFileName.Length =
            InodeFileName.MaximumLength =
                InodeFileNameLength * 2;

            InodeFileName.Buffer = FsdAllocatePool(
                NonPagedPool,
                InodeFileName.MaximumLength
                );

            FsdCharToWchar(
                InodeFileName.Buffer,
                romfs_inode->name,
                InodeFileNameLength
                );

            if (FsRtlIsNameInExpression(
                FileName,
                &InodeFileName,
                FALSE,
                NULL
                ))
            {
                *Index = FileIndex;
                Status = STATUS_SUCCESS;
                __leave;
            }

            if (InodeFileName.Buffer != NULL)
            {
                FsdFreePool(InodeFileName.Buffer);
                InodeFileName.Buffer = NULL;
            }

            FileIndex = romfs_inode->next;

        }

        Status = STATUS_NO_SUCH_FILE;
    }
    __finally
    {
        if (InodeFileName.Buffer != NULL)
        {
            FsdFreePool(InodeFileName.Buffer);
        }
    }

    return Status;
}
